@@ -7,7 +7,7 @@ class Pdb:
     prot_map = {'ALA': 'A', 'CYS': 'C', 'CYX': 'C', 'CYM': 'C', 'ASP': 'D', 'GLU': 'E', 'PHE': 'F', 'GLY': 'G',
                 'HIS': 'H', 'HIE': 'H', 'HID': 'H', 'HSD': 'H', 'HSE': 'H', 'ILE': 'I', 'LYS': 'K', 'LEU': 'L',
                 'MET': 'M', 'ASN': 'N', 'PRO': 'P', 'GLN': 'Q', 'ARG': 'R', 'SER': 'S', 'THR': 'T', 'VAL': 'V',
-                'TRP': 'W', 'TYR': 'Y', "GLUP": "E", "ASPP": "D"}
+                'TRP': 'W', 'TYR': 'Y', "GLUP": "E", "GLH": "E", "ASPP": "D", "ASH": "D", 'LYN': 'K'}
 
     nucl_map = {'DA': "A", 'DG': "G", 'DC': "C", 'DT': "T", 'DA5': "A", 'DG5': "G", 'DC5': "C", 'DT5': "T",
                 'DA3': "A", 'DG3': "G", 'DC3': "C", 'DT3': "T", 'RA': "A", 'RG': "G", 'RC': "C", 'RU': "U",
@@ -158,7 +158,7 @@ class Pdb:
     def find_missing(self):
         """
         Assuming standard naming conventions, finds atoms that are missing
-        in protein structures and prints them
+        in protein structures and prints them (heavy atoms only!)
         :return: None
         """
         pro_bb = ['N', 'O', 'C', 'CA']
@@ -231,9 +231,9 @@ class Pdb:
                     curr_resid = atom.resnum
                 if atom.resnum != curr_resid:
                     if nopbc:
-                        dist = self._atoms_dist_pbc(atom, prev_atom)
-                    else:
                         dist = self._atoms_dist(atom, prev_atom)
+                    else:
+                        dist = self._atoms_dist_pbc(atom, prev_atom)
                     prev_atom = atom
                     curr_resid = atom.resnum
                     if dist > cutoff:
@@ -316,13 +316,14 @@ class Pdb:
         Identifies and lists molecules contained in the structure, chain by chain
         :return: None
         """
+        # TODO fix
         chains = list({a.chain.strip() for a in self.atoms})
         if not chains:
             print("No chains in the molecule, run Pdb.add_chains() first")
 
         def identify(atom):
             first = atom.resname
-            if first in Pdb.prot_map.keys() or (first[1:] in Pdb.prot_map.keys() and first[0] in 'NC'):
+            if first in Pdb.prot_map.keys() or (first[1:] in Pdb.prot_map.keys() and first[0] in 'NC') or first == 'ACE':
                 return 'Protein'
             elif first in Pdb.nucl_map.keys():
                 return 'Nucleic'
@@ -336,19 +337,19 @@ class Pdb:
         for a in self.atoms:
             if mol in ['Protein', 'Nucleic']:
                 if a.chain != chain:
-                    mol_list.append([mol, 1])
+                    mol_list.append([mol + f'_chain_{chain}', 1])
                     chain = a.chain
                     mol = identify(a)
             else:
                 if a.resnum != res:
-                    if len(mol) > 1 and mol == mol_list[-1][0]:
+                    if len(mol) >= 1 and mol == mol_list[-1][0]:
                         mol_list[-1][1] += 1
                     else:
                         mol_list.append([mol, 1])
                     mol = identify(a)
                     chain = a.chain
             res = a.resnum
-        if len(mol) > 1 and mol == mol_list[-1][0] and mol not in ['Protein', 'Nucleic']:
+        if len(mol) >= 1 and mol == mol_list[-1][0]:
             mol_list[-1][1] += 1
         else:
             mol_list.append([mol, 1])
@@ -444,12 +445,13 @@ class Pdb:
         :return: None
         """
         count = offset
-        ats = self.atoms if selection is None else self.get_atom_indices(selection)
+        ats = range(len(self.atoms)) if selection is None else self.get_atom_indices(selection)
         for n in ats:
             temp = count
             try:
                 if self.atoms[n].resnum != self.atoms[n + 1].resnum or self.atoms[n].chain != self.atoms[n + 1].chain:
-                    temp = count + 1
+                    # avoid resnum==10000 as it gets written as 0 and causes problems:
+                    temp = count + 1 if not (count + 1 % 10000 == 0) else count + 2
             except IndexError:
                 pass
             self.atoms[n].resnum = count
@@ -771,8 +773,8 @@ class Pdb:
                         hook = hk
                         break
             hooksel = '{}resid {} and name {}'.format(chstr, resid, hook)
-            atomsel = '{}resid {} and name {}'.format(chstr, resid, atom_add)
             aftnr = None
+            # TODO add unittest to check we're inserting correctly
             if isinstance(aft, tuple):
                 for n, af in enumerate(aft):
                     try:
@@ -790,17 +792,14 @@ class Pdb:
             if len(geo_ref) == 2:
                 p1sel = '{}resid {} and name {}'.format(chstr, resid, geo_ref[0])
                 p2sel = '{}resid {} and name {}'.format(chstr, resid, geo_ref[1])
-                self.insert_atom(aftnr, name=atom_add, hooksel=hooksel,
+                self.insert_atom(aftnr+2, name=atom_add, hooksel=hooksel,
                                  bondlength=bond_length,
-                                 p1_sel=p1sel, p2_sel=p2sel, atomname=atom_add)
+                                 p1_sel=p1sel, p2_sel=p2sel, atomname=atom_add, resname=mutant.target_3l)
             else:
                 vec = self._vector(geo_ref, resid, chain)
-                self.insert_atom(aftnr, name=atom_add, hooksel=hooksel,
+                self.insert_atom(aftnr+2, name=atom_add, hooksel=hooksel,
                                  bondlength=bond_length,
-                                 vector=vec, atomname=atom_add)
-
-        for atom in self.get_atoms('{}resid {}'.format(chstr, resid)):
-            atom.resname = mutant.target_3l
+                                 vector=vec, atomname=atom_add, resname=mutant.target_3l)
         self.renumber_atoms()
 
     def _vector(self, atnames, resid, chain, nopbc=False):
@@ -846,21 +845,25 @@ class Pdb:
                 dists.append(self._atoms_dist_pbc(atom1, atom2))
         return dists
 
-    def check_chiral_aa(self, nopbc=False):
+    def check_chiral_aa(self, nopbc=False, fix=False):
         """
         Checks for correct chirality in amino acids, first in the backbone
-        and then for chiral side chains
+        and then for chiral side chains; to work with "fix", the last one has to be
+        a hydrogen
         :param nopbc: bool, whether to ignore PBC information
+        :param fix: bool, whether to try fixing chirality by switching the position of the hydrogen
         :return: None
         """
+        # TODO check for GLY (consistent between HA1 and HA2)?
         prot_atoms = self.get_atoms('name CA and not resname GLY')
-        self.check_chiral(prot_atoms, 'N', 'C', 'HA HA1', nopbc=nopbc)
+        self.check_chiral(prot_atoms, 'N', 'C', 'HA HA1', nopbc=nopbc, fix=fix)
         ile_atoms = self.get_atoms('name CB and resname ILE')
-        self.check_chiral(ile_atoms, 'CA', 'CG2', 'CG1', 'side chain chirality', nopbc=nopbc)
+        self.check_chiral(ile_atoms, 'CG2', 'CA', 'HB', 'side chain chirality', nopbc=nopbc, fix=fix)
         thr_atoms = self.get_atoms('name CB and resname THR')
-        self.check_chiral(thr_atoms, 'CA', 'CG1 CG2', 'OG1 OG2', 'side chain chirality', nopbc=nopbc)
+        self.check_chiral(thr_atoms, 'CG1 CG2', 'CA', 'HB', 'side chain chirality', nopbc=nopbc, fix=fix)
 
-    def check_chiral(self, cent_atoms_list, at1, at2, at3, label='backbone chirality', printing=True, nopbc=False):
+    def check_chiral(self, cent_atoms_list, at1, at2, at3, label='backbone chirality', printing=True, nopbc=False,
+                     fix=False):
         """
         Decides on correct or wrong chirality of selected chiral
         centers by calculating selected dihedrals
@@ -871,6 +874,7 @@ class Pdb:
         :param label: str, type of chirality (only to display in the warning)
         :param printing: bool, whether to print warnings (default) or return True if all checks are passed
         :param nopbc: bool, whether to ignore PBC when calculating vectors
+        :param fix: bool, whether to try fixing chirality by switching the position of the hydrogen
         :return: None (if printing), bool (if not printing)
         """
         for at in cent_atoms_list:
@@ -886,13 +890,28 @@ class Pdb:
                 else:
                     return False
             elif chi > 0:
-                if printing:
+                if fix:
+                    self.fix_chirality(h_sel=f'name {at3} and resnum {resnum} and resname {resname} {chn}',
+                                       c_sel=f'name {at.atomname} and resnum {resnum} and resname {resname} {chn}')
+                    print(f"Trying to fix chirality for residue {resname} num {resnum} by moving {at3} to the opposite "
+                          f"side of {at.atomname}, run a minimization and check again")
+                if printing and not fix:
                     print(f"Check {label} for residue {resname} num {resnum}, looks like a D-form")
-                else:
+                if not printing:
                     return False
             if not printing:
                 return True
 
+    def fix_chirality(self, h_sel, c_sel):
+        """
+        A quick-and-dirty chirality fix that moves the hydrogen to the
+        opposite side of the carbon atom (needs minimization afterwards)
+        :param h_sel: str, selection that returns the hydrogen atom to be moved
+        :param c_sel: str, selection that returns the carbon atom bound to that hydrogen
+        :return: None
+        """
+        self.reposition_atom_from_hook(h_sel, c_sel, 1.09, h_sel, c_sel)
+
     def _get_chirality(self, *atoms, nopbc=False):
         """
         Calculates a dihedral defined by 4 atoms that
@@ -968,13 +987,14 @@ class Pdb:
             assert boxline[3] == boxline[4] == boxline[6] == 0
             box = [0.0] * 6
             box[0] = boxline[0]
-            box[-1] = math.atan(boxline[1] / boxline[5]) if boxline[5] != 0 else 90.0  # TODO check twice
-            box[1] = boxline[1] / math.sin(box[-1])
+            box[-1] = math.atan(boxline[1] / boxline[5]) if boxline[5] != 0 else 90
+            box[1] = boxline[1] / math.sin(box[-1] * math.pi/180)
             box[2] = math.sqrt(boxline[7] ** 2 + boxline[8] ** 2 + boxline[2] ** 2)
             box[-2] = math.acos(boxline[7] / box[2])
-            box[-3] = math.acos((boxline[8] * math.sin(box[-1])) / box[2] + math.cos(box[-2]) * math.cos(box[-1]))
+            box[-3] = math.acos((boxline[8] * math.sin(box[-1] * math.pi/180)) / box[2]
+                                + math.cos(box[-2]) * math.cos(box[-1] * math.pi/180))
             box[0], box[1], box[2] = box[0] * 10, box[1] * 10, box[2] * 10
-            box[3], box[4], box[5] = box[3] * 180 / math.pi, box[4] * 180 / math.pi, box[5] * 180 / math.pi
+            box[3], box[4], box[5] = round(box[3] * 180/math.pi, 4), round(box[4] * 180/math.pi, 4), round(box[5], 4)
         else:
             raise RuntimeError('Can\'t read box properties')
         return atoms, tuple(box), remarks
@@ -998,12 +1018,12 @@ class Pdb:
                 selected = list(np.where(np.logical_and(dists < cutoff, dists > 0.1))[0] + 1)
                 self.conect[n + 1] = selected
 
-    def fill_beta(self, values, serials=None, smooth=False, ignore_mem=False):
+    def set_beta(self, values, selection=None, smooth=False, ignore_mem=False):
         """
         Enables user to write arbitrary values to the beta field
         of the PDB entry
         :param values: iterable; values to fill
-        :param serials: iterable, optional; can be used to specify a subset of atoms
+        :param selection: str, optional; can be used to specify a subset of atoms
         :param smooth: if float, defines sigma (in Angstrom) for beta-value smoothing
         :param ignore_mem: bool, allows to ignore memory warnings
         :return: None
@@ -1011,21 +1031,17 @@ class Pdb:
         if any([v > 999 for v in values]):
             print("At least one value is too large to fit into the `beta` field, consider division "
                   "to make values smaller")
-        if not serials:
-            serials = [a.serial for a in self.atoms]
-        if not all([i == j for i, j in zip(serials, sorted(serials))]):
-            raise ValueError("Atoms' serial numbers are not sorted, consider running .renumber_atoms() first")
-        if len(values) != len(serials):
+        atoms = self.atoms if selection is None else self.get_atoms(selection)
+        if len(values) != len(atoms):
             raise ValueError("Lists 'value' and 'serials' have inconsistent sizes: {} and {}".format(len(values),
-                                                                                                     len(serials)))
+                                                                                                     len(atoms)))
         index = 0
         if not smooth:
-            for atom in self.atoms:
-                if atom.serial in serials:
-                    atom.beta = values[index]
-                    index += 1
+            for atom in atoms:
+                atom.beta = values[index]
+                index += 1
         else:
-            if len(serials) > 10000 and not ignore_mem:
+            if len(atoms) > 10000 and not ignore_mem:
                 raise RuntimeError("Try to restrict the number of atoms (e.g. selecting CA only), or you're risking "
                                    "running out of memory. To proceed anyway, run again with ignore_mem=True")
             try:
@@ -1033,7 +1049,7 @@ class Pdb:
             except ImportError:
                 print("Needs numpy for smoothing, try installing it")
             else:
-                atomnums = [n for n, atom in enumerate(self.atoms) if atom.serial in serials]
+                atomnums = [n for n, atom in enumerate(self.atoms) if atom in atoms]
                 coords = np.array(self.get_coords())[atomnums]
                 for atom in self.atoms:
                     dists = np.linalg.norm(coords - np.array(atom.coords), axis=1)
@@ -1098,11 +1114,14 @@ class Pdb:
             for atom in self.atoms:
                 outfile.write(self._write_atom(atom, pdb=False))
             gbox = self._calc_gro_box()
-            outfile.write((9 * "{:10.5f}" + "\n").format(*gbox))
+            if sum([x**2 for x in gbox[3:]]) > 0:
+                outfile.write((9 * "{:10.5f}" + "\n").format(*gbox))
+            else:
+                outfile.write((3 * "{:10.5f}" + "\n").format(*gbox[:3]))
 
     def _calc_gro_box(self):
         if self.box[3] == self.box[4] == self.box[5] == 90.0:
-            return "{:10.5f}{:10.5f}{:10.5f}\n".format(*[x / 10 for x in self.box[:3]])
+            return [x / 10 for x in self.box[:3]] + 6 * [0.0]
         else:
             gbox = 9 * [0.0]
             conv = math.pi / 180
