@@ -293,6 +293,89 @@ class SectionMol(Section):
             for entry in subsection.entries_bonded:
                 entry.atom_numbers = tuple(n + (offset * (n >= startfrom)) for n in entry.atom_numbers)
 
+    def add_disulfide(self, resid1, resid2, other=None, rtp=None):
+        """
+        Adds a bond between the SG atoms of a selected residue pair (either within
+        or between molecules), removes the extra hydrogens, and adjusts atom types
+        or charges according to the .rtp file chosen
+        :param resid1: int, the resid of 1st residue to be linked
+        :param resid2: int, the resid of 2st residue to be linked (can be from a different molecule)
+        :param other: SectionMol (molecule), optional; if specified, the disulfide will be intermolecular
+        :param rtp: str, optional; if specified, the given .rtp will be used (otherwise interactive)
+        :return: None
+        """
+        other = self if other is None else other
+        s1 = self.get_atom(f'resid {resid1} and name SG')
+        h1 = self.get_atom(f'resid {resid1} and name HG HG1')
+        s2 = other.get_atom(f'resid {resid2} and name SG')
+        h2 = other.get_atom(f'resid {resid2} and name HG HG1')
+        assert s1.resname == s2.resname == 'CYS'
+        self.del_atom(h1.num)
+        other.del_atom(h2.num)
+        found = self.find_rtp(rtp)
+        types, charges, dihedrals, impropers, improper_type = self.parse_rtp(found)
+        disulf_names = {'CYX', 'CYS2'}
+        disulf = set([x[0] for x in types.keys()]).intersection(disulf_names)
+        if len(disulf) == 0:
+            self.top.rtp = {}
+            raise RuntimeError(f"None of the residues {disulf_names} found in the .rtp file {found}")
+        disulf_resname = disulf.pop()
+        for mol, res in zip([self, other], [resid1, resid2]):
+            for atom in mol.get_atoms(f'resid {res}'):
+                atom.type = types[(disulf_resname, atom.atomname)]
+                atom.charge = charges[(disulf_resname, atom.atomname)]
+                atom.resname = disulf_resname
+        self.merge_two(other, s1.num, s2.num)
+
+    def add_coordinated_ion(self, resid1, resid2, other=None, rtp=None):
+        """
+        Adds a bond between the cysteine sulfur or histidine nitrogen and (usually)
+        a transition metal ion like Zn or Fe, and adjusts atom types
+        or charges according to the .rtp file chosen
+        :param resid1: int, the resid of 1st residue to be linked (amino acid)
+        :param resid2: int, the resid of 2st residue to be linked (ion/heme; can be from a different molecule)
+        :param other: SectionMol (molecule), optional; if specified, the bond will be intermolecular
+        :param rtp: str, optional; if specified, the given .rtp will be used (otherwise interactive)
+        :return: None
+        """
+        other = self if other is None else other
+        c1 = self.get_atom(f'resid {resid1} and name CA')
+        if c1.resname not in {'CYS', 'HIS', 'HID', 'HIE', 'HSD', 'HSE'}:
+            raise RuntimeError(f"Residue 1 has to be cysteine or histidine, not {c1.resname}")
+        x2 = other.get_atom(f'resid {resid2} and name ZN ZN2 FE')
+        if x2.resname not in {'ZN', 'ZN2', 'FE', 'HEM', 'HEME'}:
+            raise RuntimeError(f"Residue 2 has to be a zinc or iron/heme, not {x2.resname}")
+        if c1.resname == 'CYS':
+            found = self.find_rtp(rtp)
+            types, charges, dihedrals, impropers, improper_type = self.parse_rtp(found)
+            h1 = self.get_atom(f'resid {resid1} and name HG HG1')
+            x1 = self.get_atom(f'resid {resid1} and name SG')
+            self.del_atom(h1.num)
+            sulf_names = {'CYM'}
+            sulf = set([x[0] for x in types.keys()]).intersection(sulf_names)
+            if len(sulf) == 0:
+                self.top.rtp = {}
+                raise RuntimeError(f"None of the residues {sulf_names} found in the .rtp file {found}")
+            sulf_resname = sulf.pop()
+            for atom in self.get_atoms(f'resid {resid1}'):
+                atom.type = types[(sulf_resname, atom.atomname)]
+                atom.charge = charges[(sulf_resname, atom.atomname)]
+                atom.resname = sulf_resname
+        else:
+            try:
+                self.get_atom(f'resid {resid1} and name HE2')
+            except RuntimeError:
+                try:
+                    self.get_atom(f'resid {resid1} and name HD1')
+                except RuntimeError:
+                    raise RuntimeError(f"In residue {c1.resname} {resid1}, neither HE2 nor HD1 were found, can't "
+                                       f"identify histidine protonation state")
+                else:
+                    x1 = self.get_atom(f'resid {resid1} and name NE2')
+            else:
+                x1 = self.get_atom(f'resid {resid1} and name ND1')
+        self.merge_two(other, x1.num, x2.num)
+
     def gen_state_b(self, atomname=None, resname=None, resid=None, atomtype=None, new_type=None, new_charge=None,
                     new_mass=None):
         """
@@ -686,7 +769,6 @@ class SectionMol(Section):
             # the stuff below works but is terribly ugly, we need to have API for manipulating content of Top.system
             system_setup = self.top.sections[-1].get_subsection('molecules')
             system_setup.entries = [e for e in system_setup if other.mol_name not in e]
-            self.top.recalc_sys_params()
 
     def merge_molecules(self, other):
         other.offset_numbering(self.natoms)
@@ -695,7 +777,6 @@ class SectionMol(Section):
         # the stuff below works but is terribly ugly, we need to have API for manipulating content of Top.system
         system_setup = self.top.sections[-1].get_subsection('molecules')
         system_setup.entries = [e for e in system_setup if other.mol_name not in e]
-        self.top.recalc_sys_params()
 
     def _merge_fields(self, other):
         self.top.print('WARNING watch out for #ifdef POSRES keywords that might get misplaced')
@@ -712,7 +793,11 @@ class SectionMol(Section):
     
     def _make_bond(self, atom_own, atom_other, other):
         self._get_bonds()
-        other._get_bonds()
+        try:
+            other._get_bonds()
+        except KeyError:
+            other.subsections.append(self._yield_sub([f"[ bonds ]\n"]))
+            other._get_bonds()
         new_bond = [tuple(sorted([int(atom_own), int(atom_other)]))]
         new_angles = self._generate_angles(other, atom_own, atom_other)
         new_pairs, new_dihedrals = self._generate_14(other, atom_own, atom_other)
@@ -723,6 +808,10 @@ class SectionMol(Section):
             subsection = self.get_subsection(sub)
             subsection.add_entries([gml.EntryBonded(subsection.fstring.format(*entry, subsection.prmtype), subsection)
                                     for entry in entries])
+            try:
+                other.get_subsection(sub)
+            except KeyError:
+                other.subsections.append(self._yield_sub([f"[ {sub} ]\n"]))
 
     def _remove_bond(self, at1, at2):
         self._get_bonds()
@@ -1061,16 +1150,18 @@ class SectionMol(Section):
         elif mutate_in_pdb and not self.top.pdb:
             print("No .pdb file bound to the topology, use Top.add_pdb() to add one")
 
-    def parse_rtp(self, rtp):
+    def parse_rtp(self, rtp, remember=False):
         """
         Reads an .rtp file to extract molecule definitions, separating them into
         dictionaries for: types, charges, impropers, dihedrals, bondedtypes
         :param rtp: str, path to the .rtp file
+        :param remember: bool, whether to reuse .rtps selected before
         :return: tuple of dict, each containing atom name : relevant parameter matching
         (ordered: types, charges, dihedrals, impropers, bondedtypes)
         """
         # TODO check against amber/ILDN
-        if self.top.rtp:
+        if self.top.rtp and remember:
+            print("Using previously selected .rtp file")
             return self.top.rtp['typedict'], self.top.rtp['chargedict'], self.top.rtp['dihedrals'], \
                    self.top.rtp['impropers'], self.top.rtp['bondedtypes']
         chargedict, typedict = {}, {}
@@ -1135,6 +1226,10 @@ class SectionMol(Section):
     def update_dicts(self):
         self.get_subsection('atoms').get_dicts(force_update=True)
 
+    def list_atoms(self):
+        for atom in self.atoms:
+            print(str(atom).strip())
+
     def list_bonds(self, by_types=False, by_params=False, returning=False):
         return self._list_bonded('bonds', by_types, by_params, returning)
 
@@ -1176,6 +1271,19 @@ class SectionMol(Section):
 
     def alch_h_to_ch3(self, resid, orig_name, basename, ctype=None, htype=None, ccharge=None, hcharge=0.09,
                       dummy_type='DH', add_in_pdb=True):
+        """
+        A generic routine to change a hydrogen atom into a methyl group
+        :param resid: int, ID of the residue in which the change is to be made
+        :param orig_name: str, name of the hydrogen to be modified
+        :param basename: str, an atom name for the extra atoms that will be modified with indices
+        :param ctype: str, atomtype for the methyl carbon
+        :param htype: str, atomtype for the methyl hydrogens
+        :param ccharge: float, charge for the methyl carbon
+        :param hcharge: float, charge for the methyl hydrogens
+        :param dummy_type: str, atomtype for the dummies
+        :param add_in_pdb: bool, whether to modify the associated Pdb object
+        :return: None
+        """
         if ctype is None or htype is None:
             print("Which atomtypes should be used for the methyl group:\n")
             print("[ 1 ] CT/HC (Amber methyl)")
@@ -1232,6 +1340,13 @@ class SectionMol(Section):
         self.gen_state_b(atomname=orig_name, resid=resid, new_type=dummy_type, new_charge=0, new_mass=1.008)
 
     def solute_tempering(self, temperatures, selection=None):
+        """
+        Prepares a modified topology for REST2 simulations
+        and writes the respective .top files
+        :param temperatures: list of float, solute temperatures for replica exchange
+        :param selection: str or None, can specify the selection to which the modification should be restricted
+        :return: None
+        """
         self.top.explicit_defines()
         for n, t in enumerate(temperatures):
             self.top.print(f'generating topology for effective temperature of {t} K...')
@@ -1241,6 +1356,12 @@ class SectionMol(Section):
             mod.save_top(self.top.fname.replace('.top', f'-rest{temperatures[0]/t:.3f}.top'))
 
     def scale_rest2_bonded(self, gamma, selection=None):
+        """
+        Modifies bonded terms for the REST2 protocol
+        :param gamma: float, the scaling parameter from REST2
+        :param selection: str or None, can specify the selection to which the scaling should be restricted
+        :return: None
+        """
         # get a list of atomtypes to clone
         sel = self.get_atoms(selection) if selection is not None else self.atoms
         types = {at.type for at in sel}
@@ -1262,6 +1383,12 @@ class SectionMol(Section):
             pass
 
     def scale_rest2_charges(self, gamma, selection=None):
+        """
+        Modifies charges for the REST2 protocol
+        :param gamma: float, the scaling parameter from REST2
+        :param selection: str or None, can specify the selection to which the scaling should be restricted
+        :return: None
+        """
         sel = self.get_atoms(selection) if selection is not None else self.atoms
         # scaling charges
         for a in self.atoms:
@@ -1304,6 +1431,11 @@ class SectionMol(Section):
         self.add_ff_params()
 
     def recalc_qtot(self):
+        """
+        Puts the qtot (cumulative charge of the molecule) in the comment
+        of an atom entry in the topology (done by pdb2gmx by default)
+        :return: None
+        """
         charge = 0
         for atom in self.atoms:
             charge += atom.charge
@@ -1330,6 +1462,7 @@ class SectionParam(Section):
         this fn merges them into single sections to avoid searching in all instances
         :return: None
         """
+        # TODO check for complete ifdef/endif entries, otherwise crash
         subsection_labels = [sub.label for sub in self.subsections]
         duplicated_subsections = list({label for label in subsection_labels if subsection_labels.count(label) > 1})
         if duplicated_subsections:
@@ -1359,9 +1492,15 @@ class SectionParam(Section):
                 sub.sort()  # TODO if two have same periodicity & atoms, remove one with 0-s (PMX)
 
     def find_used_ff_params(self, section='all'):
+        """
+        Finds FF parameters that are used by the system
+        :param section: str, for which section should the search be performed ('all' or 'atomtypes', '...')
+        :return: list of EntryParams that are used by the system's topology
+        """
         used_params = []
         if section == 'all':
             subsections_to_add = ['atomtypes', 'pairtypes', 'nonbonded_params', 'constrainttypes']
+            # TODO check what with other sections?
         else:
             subsections_to_add = [section]
         for sub in subsections_to_add:
@@ -1374,6 +1513,20 @@ class SectionParam(Section):
                     used_params.extend(ssub.find_used_ff_params())
         return used_params
 
+    def fix_zero_periodicity(self):
+        """
+        Recent versions of OpenMM require non-zero dihedral periodicities,
+        so this function changes 0s to 1s (assuming the force constant is 0 too)
+        :return: None
+        """
+        subsections = self.get_subsections('dihedraltypes')
+        for sub in subsections:
+            if str(sub.prmtype) in '149':
+                for entry in sub.entries_param:
+                    if entry.params[-1] == 0:
+                        assert entry.params[-2] == 0
+                        entry.params[-1] = 1
+
     def clone_type(self, atomtype, prefix='y', new_type=None):
         """
         Generates an exact type of a selected atomtype,
@@ -1399,6 +1552,12 @@ class SectionParam(Section):
         self._remove_symm_dupl(new_atomtype)
 
     def clean_unused(self, used_params, section='all'):
+        """
+        Cleans up FF parameters that are not used by the given system
+        :param used_params: list of str, identifiers of parameters we already know are being used
+        :param section: str, which sections to clean up ('all' for all of them or 'bonds', 'angles', 'dihedrals', etc.)
+        :return: None
+        """
         matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes',
                      'impropers': 'dihedraltypes', 'atomtypes': 'atomtypes', 'pairtypes': 'pairtypes',
                      'nonbonded_params': 'nonbonded_params', 'constrainttypes': 'constrainttypes'}
@@ -1444,6 +1603,16 @@ class SectionParam(Section):
         ss.set_opt_dih(values)
 
     def add_nbfix(self, type1, type2, mod_sigma=0.0, mod_epsilon=0.0, action_default='x'):
+        """
+        Generates NBFIX entries for the chosen pair of atomtypes by modifying the current
+        Lorentz-Berthelot rules, or an existing NBFIX
+        :param type1: str, name of the 1st atomtype in the pair
+        :param type2: str, name of the 2nd atomtype in the pair
+        :param mod_sigma: float, by how much to increase the LJ sigma (in nm)
+        :param mod_epsilon: float, by how much to increase the LJ epsilon (in kJ/mol)
+        :param action_default: str, what to do if an NBFIX already exists (check prompt if that happens)
+        :return: None
+        """
         atp = self.get_subsection('atomtypes')
         sigma1, eps1, sigma2, eps2 = [None] * 4
         for entry in atp:
@@ -1483,6 +1652,15 @@ class SectionParam(Section):
 
     @staticmethod
     def gen_clones(entry, atomtype, new_atomtype):
+        """
+        Copies entry that contains an atomtype to be cloned, taking into
+        account possible multiple occurences of that type and generating
+        all possible combinations/permutations of the changes
+        :param entry: an EntryParam instance, the entry to be copied
+        :param atomtype: str, the atomtype that is being cloned
+        :param new_atomtype: str, the new atomtype being created
+        :return: list of str, new lines to be added
+        """
         lines = []
         nchanges = entry.types.count(atomtype)
         changes = []
@@ -1498,6 +1676,15 @@ class SectionParam(Section):
 
     @staticmethod
     def mod_types(entry, mods, new_atomtype, atomtype):
+        """
+        Modifies an EntryParam to include the new atomtype instead of the old one
+        in places specified by the directive in mods
+        :param entry: an EntryParam instance, the entry to be copied
+        :param mods: list of lists, specifies where to put the modification
+        :param new_atomtype: str, the new atomtype being created
+        :param atomtype: str, the atomtype that is being cloned
+        :return: str, the cloned line
+        """
         line = str(entry)
         lentype = len(atomtype)
         for num in mods[::-1]:
